"""
Copyright 2008-2014 by Aerospike, Inc. All rights reserved.
THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE. THE COPYRIGHT NOTICE
ABOVE DOES NOT EVIDENCE ANY ACTUAL OR INTENDED PUBLICATION.
"""


# -*- coding: utf-8 -*-
#from libs import base
#from constants import *
#from utilities import connection_manager
from libs import base
from constants import *
from utilities import connection_manager
from helpers.debugger import Debugger
from distutils import version
import Queue

#from flaskapp import app
#logger = app.logger

"""
NOTE: Currently XDR is not secured. Once security is enabled for XDR port
pass on proper credentials to base.
"""

debugger = Debugger()

class Xdr(object):
    """Xdr contains all information about xdr"""

    def __new__(cls, ip, port, infoport, cluster_id, build, user=None, password=None):

        xdr_enabled = False
        # Check for XDR in ASD
        features = base.get_node_features(
            ip, port, user=user, password=password)

        xdr_in_asd = False
        if 'xdr' in features:
            xdr_in_asd = True

        if xdr_in_asd:
            config = base.get_node_configuration(ip, port, user, password)
            if not config.get('node_status', None) == 'off':
                # There are two enable-xdr, convertor returns both values
                # comma separated
                enable_xdr = config.get('enable-xdr', None)
                if enable_xdr and enable_xdr.split(",")[0] == 'true':
                    xdr_enabled = True

        # Check if XDR listening on default XDR port
        if not xdr_enabled:

            if not xdr_in_asd:
                try:
                    if int(infoport) == int(port):
                        return {}
                except ValueError:
                    return {}

            if user:
                xdr_enabled = base.is_xdr_enabled(
                    ip, infoport, user, password)
                if not xdr_enabled:
                    xdr_enabled = base.is_xdr_enabled(
                        ip, infoport)
            else:
                xdr_enabled = base.is_xdr_enabled(
                    ip, infoport)

        if xdr_enabled:
            return super(Xdr, cls).__new__(
                cls, ip, port, infoport, cluster_id,
                user=user, password=password)
        else:
            return {}

    def __init__(
            self, ip, port, infoport, cluster_id, build, user=None, password=None):

        self._ip = ip
        self._port = port
        self._infoport = infoport
        self._cluster_id = cluster_id
        self._build = build
        self._statistics = {}
        self._raw_statistics = {}
        self._modified_statistics = {}
        self._modified_raw_statistics = {}
        self._configuration = {}
        self._modified_configuration = {}
        self._dc_statistics = {}

        self._xdr_in_asd = False
        self._xdr_infoport_enabled = False
        self._xdr_infoport_secured = False

        # Check for XDR in ASD
        features = base.get_node_features(
            ip, port, user=user, password=password)
        if 'xdr' in features:
            self._xdr_in_asd = True
            config = base.get_node_configuration(ip, port, user, password)
            if not config.get('node_status', None) == 'off':
                enable_xdr = config.get('enable-xdr', None)
                # There are two enable-xdr as of AS 3.7.0.2, convertor returns
                # both values comma separated
                if enable_xdr and enable_xdr.split(",")[0] == 'true':
                    self._xdr_infoport_enabled = base.is_xdr_enabled(
                        ip, infoport, user, password)

        # Check if XDR default infoport is enabled
        if not self._xdr_in_asd and not self._xdr_infoport_enabled:
            if user:
                self._xdr_infoport_enabled = base.is_xdr_enabled(
                    ip, infoport, user, password)
                if not self._xdr_infoport_enabled:
                    self._xdr_infoport_enabled = base.is_xdr_enabled(
                        ip, infoport)
                else:
                    self._xdr_infoport_secured = True
            else:
                self._xdr_infoport_enabled = base.is_xdr_enabled(
                    ip, infoport)

        if self._xdr_in_asd:
            self._asd_socket_queue = Queue.Queue(maxsize=SOCKET_QUEUE_SIZE)
            self._asd_socket_queue = connection_manager.create_conn_queue(
                ip, port, int(XDR_SOCKET_QUEUE_SIZE), user=user,
                password=password)

        if self._xdr_infoport_enabled:
            self._socket_queue = Queue.Queue(maxsize=SOCKET_QUEUE_SIZE)
            if self._xdr_infoport_secured:
                self.close_sockets()
                self._socket_queue = connection_manager.create_conn_queue(
                    ip, infoport, int(XDR_SOCKET_QUEUE_SIZE),
                    user=user, password=password)
            else:
                self.close_sockets()
                self._socket_queue = connection_manager.create_conn_queue(
                    ip, infoport, int(XDR_SOCKET_QUEUE_SIZE))

    def get_socket_from_queue(self):

        while self._socket_queue.empty():
            pass
        sock = self._socket_queue.get()
        #return self._socket_queue.get()
        return sock

    def put_socket_in_queue(self, sock):

        #import sys, inspect
        self._socket_queue.put(sock)

    # ASD Socket Queue
    def get_socket_from_asd_queue(self):
        while self._asd_socket_queue.empty():
            pass
        sock = self._asd_socket_queue.get()
        return sock

    def put_socket_in_asd_queue(self, sock):
        self._asd_socket_queue.put(sock)

    def get_statistics(self, stats_type="all", raw=False, user=None, password=None):

        # Try with ASD port
        if self._xdr_in_asd:
            asd_sock = self.get_socket_from_asd_queue()
            statistics = base.get_xdr_statistics(
                self._ip, self._port, user=None, password=None, sock=asd_sock)
            if statistics.get('node_status', None) is 'off':
                new_sock = connection_manager.create_socket(
                    self._ip, self._port, user=user, password=password)
                if new_sock is not None:
                    asd_sock = new_sock
                    statistics = base.get_xdr_statistics(
                        self._ip, self._port, user=user, password=password,
                        sock=asd_sock)
            self.put_socket_in_asd_queue(asd_sock)

        # Try default XDR port if XDR not in ASD
        if not self._xdr_in_asd or statistics.get('node_status', None) is 'off':
            if self._xdr_infoport_enabled:
                sock = self.get_socket_from_queue()
                if self._xdr_infoport_secured:
                    statistics = base.get_xdr_statistics(
                        self._ip, self._infoport,
                        user=user, password=password, sock=sock)
                else:
                    statistics = base.get_xdr_statistics(
                        self._ip, self._infoport, sock=sock)

                if statistics.get('node_status', None) is 'off':
                    if self._xdr_infoport_secured:
                        new_sock = connection_manager.create_socket(
                            self._ip, self._infoport,
                            user=user, password=password)
                    else:
                        new_sock = connection_manager.create_socket(
                            self._ip, self._infoport)

                    if new_sock is not None:
                        sock = new_sock
                        if self._xdr_infoport_secured:
                            statistics = base.get_xdr_statistics(
                                self._ip, self._infoport,
                                user=user, password=password, sock=sock)
                        else:
                            statistics = base.get_xdr_statistics(
                                self._ip, self._infoport, sock=sock)
                self.put_socket_in_queue(sock)

        raw_statistics = statistics.copy()

        # for version 3.9 and after
        if version.LooseVersion(self._build) >= version.LooseVersion("3.9.0"):
            statistics = self.assign_stat_aliases(statistics)

        self._modified_statistics = dict(set(statistics.items()) - set(self._statistics.items()))
        self._modified_raw_statistics = dict(set(raw_statistics.items()) - set(self._raw_statistics.items()))
        self._statistics = statistics
        self._raw_statistics = raw_statistics

        if raw:
            return self._raw_statistics if stats_type == "all" else self._modified_raw_statistics
        else:
            return self._statistics if stats_type == "all" else self._modified_statistics

    def assign_stat_aliases(self, stats):
        stats['free_dlog_pct'] = stats.get("dlog_free_pct", None)
        stats['stat_recs_logged'] = stats.get("dlog_logged", None)
        stats['stat_dlog_recs_overwritten'] = stats.get("dlog_overwritten_error", None)
        stats['stat_recs_linkdown_processed'] = stats.get("dlog_processed_link_down", None)
        stats['stat_recs_localprocessed'] = stats.get("dlog_processed_main", None)
        stats['stat_recs_replprocessed'] = stats.get("dlog_processed_replica", None)
        stats['stat_recs_relogged'] = stats.get("dlog_relogged", None)
        stats['used_recs_dlog'] = stats.get("dlog_used_objects", None)
        stats['local_recs_notfound'] = stats.get("xdr_read_not_found", None)
        stats['failednode_sessions_pending'] = stats.get("xdr_active_failed_node_sessions", None)
        stats['linkdown_sessions_pending'] = stats.get("xdr_active_link_down_sessions", None)
        stats['esmt_ship_avg_comp_pct'] = stats.get("xdr_ship_compression_avg_pct", None)
        stats['hotkeys_fetched'] = stats.get("xdr_hotkey_fetch", None)
        stats['noship_recs_hotkey'] = stats.get("xdr_hotkey_skip", None)
        stats['stat_recs_inflight'] = stats.get("xdr_ship_inflight_objects", None)
        stats['stat_recs_outstanding'] = stats.get("xdr_ship_outstanding_objects", None)
        stats['err_recs_dropped'] = stats.get("xdr_queue_overflow_error", None)
        stats['read_threads_avg_processing_time_pct'] = stats.get("xdr_read_active_avg_pct", None)
        stats['local_recs_error'] = stats.get("xdr_read_error", None)
        stats['read_threads_avg_waiting_time_pct'] = stats.get("xdr_read_idle_avg_pct", None)
        stats['local_recs_fetch_avg_latency'] = stats.get("xdr_read_latency_avg", None)
        stats['dispatch_request_queue_used_pct'] = stats.get("xdr_read_reqq_used_pct", None)
        stats['dispatch_request_queue_used'] = stats.get("xdr_read_reqq_used", None)
        stats['dispatch_response_queue_used'] = stats.get("xdr_read_respq_used", None)
        stats['local_recs_fetched'] = stats.get("xdr_read_success", None)
        stats['transaction_queue_used_pct'] = stats.get("xdr_read_txnq_used_pct", None)
        stats['transaction_queue_used'] = stats.get("xdr_read_txnq_used", None)
        stats['stat_recs_relogged_incoming'] = stats.get("xdr_relogged_incoming", None)
        stats['stat_recs_relogged_outgoing'] = stats.get("xdr_relogged_outgoing", None)
        stats['esmt_bytes_shipped'] = stats.get("xdr_ship_bytes", None)
        stats['xdr_deletes_shipped'] = stats.get("xdr_ship_delete_success", None)
        stats['err_ship_server'] = stats.get("xdr_ship_destination_error", None)
        stats['latency_avg_ship'] = stats.get("xdr_ship_latency_avg", None)
        stats['err_ship_client'] = stats.get("xdr_ship_source_error", None)
        stats['stat_recs_shipped_ok'] = stats.get("xdr_ship_success", None)
        stats['stat_recs_shipped'] = stats.get("xdr_ship_success", None)
        stats['cur_throughput'] = stats.get("xdr_throughput", None)
        stats['noship_recs_uninitialized_destination'] = stats.get("xdr_uninitialized_destination_error", None)
        stats['noship_recs_unknown_namespace'] = stats.get("xdr_unknown_namespace_error", None)

        return stats

    def get_subset_of_statistics(self, attrs=[], user=None, password=None):

        subset = dict()
        stats = self.get_statistics(user=user, password=password)
        for attr in attrs:
            subset[attr] = stats.get(attr, NOT_SUPPORTED)
        status = stats.get('node_status')
        return subset, status

    def get_statistic(self, attr, user=None, password=None):

        stats = self.get_statistics(user=user, password=password)
        return stats.get(attr, NOT_SUPPORTED)

    def get_configurations(self, stats_type="all", user=None, password=None):

        if self._xdr_in_asd:
            asd_sock = self.get_socket_from_asd_queue()
            configuration = base.get_node_configuration(
                self._ip, self._port, user=user, password=password,
                sock=asd_sock)
            self.put_socket_in_asd_queue(asd_sock)

        if not self._xdr_in_asd or configuration.get('node_status', None) \
                is 'off':
            sock = self.get_socket_from_queue()
            if self._xdr_infoport_secured:
                configuration = base.get_node_configuration(
                    self._ip, self._infoport, user=user,
                    password=password, sock=sock)
            else:
                configuration = base.get_node_configuration(
                    self._ip, self._infoport, sock=sock)
            self.put_socket_in_queue(sock)

        # sock = self.get_socket_from_queue()
        # configuration = base.get_node_configuration(self._ip, self._infoport, sock=sock)
        # #if configuration.get('node_status',None) is 'off':
        #     #new_sock = connection_manager.create_socket(self._ip. self._infoport, user=user, password=password)
        #     #if new_sock is not None:
        #      #   sock = new_sock
        #       #  configuration = base.get_node_configuration(self._ip, self._infoport, sock=sock)
        # self.put_socket_in_queue(sock)
        self._modified_configuration = dict(set(configuration.items()) - set(self._configuration.items()))
        self._configuration = configuration
        return self._configuration if stats_type == "all" else self._modified_configuration

    def get_subset_of_configurations(self, attrs=[], user=None, password=None):

        subset = dict()
        configurations = self.get_configurations(user=user, password=password)

        for attr in attrs:
            subset[attr] = configurations.get(attr, NOT_SUPPORTED)
        return subset

    def get_configuration(self, attr, user=None, password=None):

        configurations = self.get_configurations(user=user, password=password)
        return configurations.get(attr, NOT_SUPPORTED)

    def update_configurations(self, user=None, password=None):

        if self._xdr_in_asd:
            asd_sock = self.get_socket_from_asd_queue()
            configuration = base.get_node_configuration(
                # self._ip, self._port, user=user, password=password, clear_cache=True, sock=asd_sock)
                self._ip, self._port, user=user, password=password, sock=asd_sock)
            self.put_socket_in_asd_queue(asd_sock)

        if not self._xdr_in_asd or configuration.get('node_status', None) \
                is 'off':
            sock = self.get_socket_from_queue()
            if self._xdr_infoport_secured:
                configuration = base.get_node_configuration(
                    # self._ip, self._infoport, user=user, password=password, clear_cache=True, sock=sock)
                    self._ip, self._infoport, user=user, password=password, sock=sock)
            else:
                # configuration = base.get_node_configuration(self._ip, self._infoport, clear_cache=True, sock=sock)
                configuration = base.get_node_configuration(self._ip, self._infoport, sock=sock)
            self.put_socket_in_queue(sock)

        # sock = self.get_socket_from_queue()
        # configuration = base.get_node_configuration(self._ip, self._infoport, clear_cache=True, sock=sock)
        # #if configuration.get('node_status', None) is 'off':
        #     #new_sock = connection_manager.create_socket(self._ip. self._infoport, user=user, password=password)
        #     #if new_sock is not None:
        #     #    sock = new_sock
        #     #    configuration = base.get_node_configuration(self._ip, self._infoport, clear_cache=True, sock=sock)
        # self.put_socket_in_queue(sock)
        self._modified_configuration = dict(set(configuration.items()) - set(self._configuration.items()))
        self._configuration = configuration

    def set_configuration(self, parameter, value, user=None, password=None):

        if self._xdr_in_asd:
            asd_sock = self.get_socket_from_asd_queue()
            status = base.set_xdr_configuration(
                self._ip, self._port, str(parameter), str(value),
                user=user, password=password, sock=asd_sock)
            self.put_socket_in_asd_queue(asd_sock)

        if not self._xdr_in_asd or not status:
            sock = self.get_socket_from_queue()
            if self._xdr_infoport_secured:
                status = base.set_xdr_configuration(
                    self._ip, self._infoport, str(parameter), str(value),
                    user=user, password=password, sock=sock)
            else:
                status = base.set_xdr_configuration(
                    self._ip, self._infoport, str(parameter), str(value),
                    sock=sock)
            self.put_socket_in_queue(sock)

        # sock = self.get_socket_from_queue()
        # status = base.set_xdr_configuration(self._ip, self._infoport, str(parameter), str(value), sock=sock)
        # #if status is False:
        #  #   new_sock = connection_manager.create_socket(self._ip,self._infoport, user=user, password=password)
        #   #  if new_sock is not None:
        #    #     sock = new_sock
        #     #    status = base.set_xdr_configuration(self._ip, self._infoport, str(parameter), str(value), sock=sock)
        # self.put_socket_in_queue(sock)
        return status

    def set_configurations(self, edit_parameters, user=None, password=None):

        error_list = list()
        for key, value in list(edit_parameters.items()):
            if self.set_configuration(key, value, user=user, password=password) is False:
                error_list.append(key)
        self.update_configurations(user=user, password=password)

        return error_list

    def get_dc_statistics(self, dc, user=None, password=None):

        if self._xdr_in_asd:
            asd_sock = self.get_socket_from_asd_queue()
            self._dc_statistics = base.get_dc_statistics(
                self._ip, self._port, dc, user=user, password=password,
                sock=asd_sock)
            self.put_socket_in_asd_queue(asd_sock)

        if not self._xdr_in_asd or not self._dc_statistics:
            sock = self.get_socket_from_queue()
            if self._xdr_infoport_secured:
                self._dc_statistics = base.get_dc_statistics(
                    self._ip, self._infoport, dc, user=user, password=password,
                    sock=sock)
            else:
                self._dc_statistics = base.get_dc_statistics(
                    self._ip, self._infoport, dc, sock=sock)
            self.put_socket_in_queue(sock)

        #sock = self.get_socket_from_queue()
        #self._dc_statistics = base.get_dc_statistics(self._ip, self._infoport, dc, sock=sock)
        # self._dc_statistics = base.get_dc_statistics(self._ip, self._infoport, dc)
        """if self._dc_statistics is False:
            new_sock = connection_manager.create_socket(self._ip, self._infoport, user=user, password=password)
            if new_sock is not None:
                sock = new_sock
                self._dc_statistics = base.get_dc_statistics(self._ip, self._infoport, dc, sock=sock)"""
        #self.put_socket_in_queue(sock)
        return self._dc_statistics

    def get_subset_of_dc_statistics(self, attrs=[], user=None, password=None):

        """
        This method currently not in USE
        """
        subset = dict()

        stats = self.get_dc_statistics(user=user, password=password)
        for attr in attrs:
            subset[attr] = stats.get(attr, NOT_SUPPORTED)
        return subset

    def get_dc_statistic(self, attr, user=None, password=None):

        """
        This method currently not in USE
        """
        stats = self.get_dc_statistics(user=user, password=password)
        return stats.get(attr, NOT_SUPPORTED)

    def get_dc_names(self, user=None, password=None):

        if self._xdr_in_asd:
            asd_sock = self.get_socket_from_asd_queue()
            names = base.get_dc_names(
                self._ip, self._port, user=user, password=password, sock=asd_sock)
            self.put_socket_in_asd_queue(asd_sock)

        if not self._xdr_in_asd or type(names) is int:
            sock = self.get_socket_from_queue()
            if self._xdr_infoport_secured:
                names = base.get_dc_names(
                    self._ip, self._infoport,
                    user=user, password=password, sock=sock)
            else:
                names = base.get_dc_names(
                    self._ip, self._infoport,
                    user=user, password=password, sock=sock)
            self.put_socket_in_queue(sock)

        # sock = self.get_socket_from_queue()
        # names = base.get_dc_names(self._ip, self._infoport, sock=sock)
        # #if type(names) is int:
        #  #   new_sock = connection_manager.create_socket(self._ip, self._infoport, user=user, password=password)
        #   #  if new_Sock is not None:
        #     #    sock = new_sock
        #    #     names = base.get_dc_names(self._ip, self._infoport, sock=sock)
        # self.put_socket_in_queue(sock)
        return names

    def get_xdr_port(self):

        return self._infoport

    def replace_socket_queue(self, user, password):
        if self._xdr_infoport_secured:
            self._socket_queue = connection_manager.create_conn_queue(
                self._ip, self._infoport, int(XDR_SOCKET_QUEUE_SIZE),
                user=user, password=password)
        else:
            self._socket_queue = connection_manager.create_conn_queue(
                self._ip, self._infoport, int(XDR_SOCKET_QUEUE_SIZE))

    def get_socket_queue(self):

        try:
            return self._socket_queue
        except AttributeError:
            return Queue.Queue(maxsize=SOCKET_QUEUE_SIZE)

    def close_sockets(self):

        xdr_queue = self.get_socket_queue()
        connection_manager.close_sockets_in_queue(xdr_queue)

    def get_dc_config(self, user=None, password=None):
        response = {}
        if self._xdr_in_asd:
            asd_sock = self.get_socket_from_asd_queue()
            response = base.get_dc_config(
                self._ip, self._port, user=user,
                password=password, sock=asd_sock
            )
            if not response:
                new_sock = connection_manager.create_socket(
                    self._ip, self._port, user=user, password=password)
                if new_sock is not None:
                    asd_sock = new_sock
                    response = base.get_dc_config(
                        self._ip, self._infoport, user=user,
                        password=password, sock=asd_sock
                    )
            self.put_socket_in_asd_queue(asd_sock)

        if not self._xdr_in_asd or not response:
            if self._xdr_infoport_enabled:
                sock = self.get_socket_from_queue()
                if self._xdr_infoport_secured:
                    response = base.get_dc_config(
                        self._ip, self._infoport, user=user,
                        password=password, sock=sock
                    )
                else:
                    response = base.get_dc_config(
                        self._ip, self._infoport, sock=sock)
                if not response:
                    if self._xdr_infoport_secured:
                        new_sock = connection_manager.create_socket(
                            self._ip, self._infoport,
                            user=user, password=password)
                    else:
                        new_sock = connection_manager.create_socket(
                            self._ip, self._infoport)
                    if new_sock is not None:
                        sock = new_sock
                        if self._xdr_infoport_secured:
                            response = base.get_dc_config(
                                self._ip, self._infoport, user=user,
                                password=password, sock=sock)
                        else:
                            response = base.get_dc_config(
                                self._ip, self._infoport, sock=sock)
                self.put_socket_in_queue(sock)
        return response

    def get_debug_status(self):
        return debugger.status()
